/* 
	GameGrid.java
	Jason C. Judt
	This file was writen for the ICFP 2002 contest and may be used by anyone, anywhere
	without restriction. Enjoy. 
*/

package myprojects.jbotchallenge;

import java.awt.Point;
import java.util.*;

public class GameGrid extends Object
{
	// grid cell type
	public static final int OPEN_SPACE  = 0;
	public static final int WALL		= 1;
	public static final int WATER		= 2;
	public static final int HOME_BASE	= 3;
	
	// direction restrictions
	public static final int NONE  	= 0;
	public static final int NORTH	= 1;
	public static final int SOUTH	= 2;
	public static final int EAST	= 3;
	public static final int WEST	= 4;
	
	// maybe change factors to ints, keep track of worst areas, 
	// use the maximum value on the board to normalize the rest
	
	// static repulsing factors
	public static final double WATER_0_FACTOR 		= 1.0;
	public static final double WATER_1_FACTOR 		= 0.5;
	public static final double WATER_2_FACTOR 		= 0.125;
	public static final double WALL_0_FACTOR		= 0.90;
	public static final double WALL_1_FACTOR		= 0.225;
	public static final double WALL_2_FACTOR		= 0.056;
	public static final double HOMEBASE_0_FACTOR 	= 0.012;
	public static final double HOMEBASE_1_FACTOR 	= 0.003;
	public static final double HOMEBASE_2_FACTOR 	= 0.001;
	
	// dynamic repulsing factors
	public static final double BOT_0_FACTOR 		= 0.5;
	public static final double BOT_1_FACTOR 		= 0.125;
	public static final double BOT_2_FACTOR			= 0.031;
	public static final double BASE_0_FACTOR 		= 0.2;
	public static final double BASE_1_FACTOR		= 0.05;
	public static final double BASE_2_FACTOR		= 0.001;
	
	// static attracting factors
	public static final double SPACE_0_FACTOR		= -0.036;
	public static final double SPACE_1_FACTOR		= -0.009;
	public static final double SPACE_2_FACTOR		= -0.002;
	
	// dynamic attracting factors
	public static final double DROPZONE_0_FACTOR	= -1.0;
	public static final double DROPZONE_1_FACTOR	= -0.5;
	public static final double DROPZONE_2_FACTOR	= -0.125;
	
	class GridCell
	{
		
		public GridCell(int x, int y)
		{
			type = OPEN_SPACE;
			repulstion = 0;	
			directionRestriction = NONE;
			bots = new Vector();
			packages = new Vector();
			canGoNorth = true;
			canGoSouth = true;
			canGoEast = true;
			canGoWest = true;
			
			hopCount = -1;
			parentDirection = NONE;
			this.x = x;
			this.y = y;
		}
		
		public int 		type;
		public double 	repulstion;
		public int 		directionRestriction;
		public Vector 	bots;
		public Vector	packages;
		public boolean	canGoNorth;
		public boolean 	canGoSouth;
		public boolean 	canGoEast;
		public boolean 	canGoWest;
		
		// for pathfinding
		public int 		hopCount;
		public int 		parentDirection;
		public int 		x;
		public int 		y;
		
	}
	
	public GameGrid()
	{
		m_PackageCatalog = new Vector();
		m_RobotCatalog = new Vector();
		m_HomeBases	 = new Vector();	
	}
	
	public void initialize(int sizeX, int sizeY)
	{
		m_RowCountX = sizeX;
		m_ColCountY = sizeY;
		m_Rows = new GridCell[m_RowCountX][m_ColCountY];
		for (int i=0; i<m_RowCountX; i++)
		{
			for (int j=0; j<m_ColCountY; j++)
			{
				m_Rows[i][j] = new GridCell(i, j);	
			}	
		}
		System.out.println("Grid Size - " + Integer.toString(sizeX) + " x " + Integer.toString(sizeY)); 
		
	}
	
	public int getCellTypeAt(int x, int y)
	{
		return m_Rows[x-1][y-1].type; 	
	}
	
	public int getRowCount()
	{
		return m_RowCountX;
	}
	
	public int getColCount()
	{
		return m_ColCountY;	
	}
	
	public Vector getCurrentPackage()
	{
		return m_CurrentPackages;
	}
	
	public Vector getPackageCatalog()
	{
		return m_PackageCatalog;	
	}
	
	public Vector getHomeBases()
	{
		return m_HomeBases;
	}
	public void setCurrentPackages(Vector p)
	{
		m_CurrentPackages = p;
		
		// lets add these to the catalog
		
		for (int i=0; i<m_CurrentPackages.size(); i++)
		{
			PackageObject po = (PackageObject)m_CurrentPackages.elementAt(i);
			po.setLocation(m_OurBot.getLocationX(), m_OurBot.getLocationY());
			po.setState(PackageObject.DROPPED_STATE);
		
			Iterator it = m_PackageCatalog.iterator();
			while (it.hasNext())
			{
				PackageObject poTemp = (PackageObject) it.next();
				if (poTemp.getID() == po.getID())
				{
					it.remove();
				}			
			}
			m_PackageCatalog.add(po);		
		}
		
		// also remove items from catalog if they were delivered!!!
		Iterator it2 = m_PackageCatalog.iterator();
		while (it2.hasNext())
		{
			PackageObject poTemp2 = (PackageObject) it2.next();
			if (poTemp2.getState() == PackageObject.DROPPED_STATE)
			{
				if (poTemp2.getLocationX() == m_OurBot.getLocationX())
				{
					if (poTemp2.getLocationY() == m_OurBot.getLocationY())
					{
						if (!(m_CurrentPackages.contains(poTemp2)))
						{
							it2.remove();
							m_OurBot.getPackages().remove(poTemp2);
						}
					}
				}
			}
		}	
	}
	
	public void readRow(String rowData, int colYNumber)
	{
		char [] rowChars = rowData.toCharArray();
//		System.out.println("Row #" + Integer.toString(colYNumber) + " is - " + Integer.toString(rowChars.length));
		for (int i=0; i<rowChars.length; i++)
		{
//			System.out.println("char #" + Integer.toString(i));
			char ch = rowChars[i];
			if (ch == '.')
				m_Rows[i][colYNumber - 1].type = OPEN_SPACE;
			if (ch == '~')
				m_Rows[i][colYNumber - 1].type = WATER;
			if (ch == '#')
				m_Rows[i][colYNumber - 1].type = WALL;
			if (ch == '@')
			{
				m_Rows[i][colYNumber - 1].type = HOME_BASE;
				// change to coordinates from array indexes
				m_HomeBases.add(new Point(i + 1, colYNumber));  
			}
		}
	}
	
	public void preprocessGrid()
	{
		// assign restrictions
		assignRestrictions();
		// update gradients	
	}
	
	public void assignRestrictions()
	{
		for (int i=0; i<m_ColCountY; i++)
		{
			for (int j=0; j<m_RowCountX; j++)
			{
				
				// first, factor boarders
				if (j == 0) 	// if we are on west rim
				{
					m_Rows[j][i].canGoWest = false;
				}
				
				if (j == m_RowCountX - 1)	// if we are on east rim
				{
					m_Rows[j][i].canGoEast = false;
				}
				
				if (i == 0)		// if we are on north rim
				{
					m_Rows[j][i].canGoSouth = false;
				}
				
				if (i == m_ColCountY - 1) // if we are on south rim
				{
					m_Rows[j][i].canGoNorth = false;
				}
				
				// next process water  and wall restrictions
				if ((m_Rows[j][i].type == WATER) || (m_Rows[j][i].type == WALL))
				{
					// west block can't go east
					if (j > 0)
						m_Rows[j-1][i].canGoEast = false;
					// east block can't go west
					if (j < m_RowCountX - 1)
						m_Rows[j+1][i].canGoWest = false;
					// south block can't go north
					if (i > 0)
						m_Rows[j][i-1].canGoNorth = false;
					// north block can't go south
					if (i < m_ColCountY - 1)
						m_Rows[j][i+1].canGoSouth = false;
				}
				
			}
		}
	}
	
	public void updateGradients()
	{
		// set initial values
		for (int i=0; i<m_RowCountX; i++)
		{
			for (int j=0; j<m_ColCountY; j++)
			{
				
				// first, factor boarders
				if ((j == 0) || (j == m_ColCountY - 1) || (i == 0) || (i == m_RowCountX - 1))
				{
					
				}
				
			}
		}
				
	}
	
	public Vector getRobotsWithinTwoStepsOfPoint(int originX, int originY)
	{
		Vector robots = new Vector();
		resetHopCount();
		GridCell origin = m_Rows[originX - 1][originY - 1];
		int depth = 0;
		// begin search
		GridCell currentCell = origin;
		currentCell.hopCount = 0;
		GridCell nextCell = null;
		Vector holder = new Vector();
		Vector newV = null;
		Vector oldV = new Vector();
		oldV.add(currentCell);
		holder.add(oldV);
		while (depth <= 2)
		{
			newV = new Vector();
			for (int i=0; i<oldV.size(); i++)
			{
				currentCell = (GridCell)oldV.get(i);
				if (currentCell.canGoNorth)
				{
					nextCell = m_Rows[currentCell.x][currentCell.y + 1];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = SOUTH;
						nextCell.hopCount = depth;
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");
					}
				}
				if (currentCell.canGoEast)
				{
					nextCell = m_Rows[currentCell.x + 1][currentCell.y];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = WEST;
						nextCell.hopCount = depth;
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");	
						
					}
				}
				if (currentCell.canGoSouth)
				{
					nextCell = m_Rows[currentCell.x][currentCell.y - 1];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = NORTH;
						nextCell.hopCount = depth;
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");
					}
				}
				if (currentCell.canGoWest)
				{
					nextCell = m_Rows[currentCell.x - 1][currentCell.y];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = EAST;
						nextCell.hopCount = depth;
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");
					}
				}
			}
		
			oldV = newV;
			holder.add(oldV);
			depth++;
		}
		
		// so we got two layers of nodes after the origin...let's iterate through them
		// and see if any of them have robots
		for (int i=0; i<holder.size(); i++)
		{
			Vector layer = (Vector) holder.get(i);
			for (int j=0; j<layer.size(); j++)
			{
				GridCell gc = (GridCell) layer.get(j);
				int x = gc.x + 1;
				int y = gc.y + 1;
				for (int k=0; k<m_RobotCatalog.size(); k++)
				{
					Robot r = (Robot) m_RobotCatalog.get(k);
					if (r.getLocationX() == x)
					{
						if (r.getLocationY() == y)
						{
							robots.add(r);	
						}
					}
				}
			}
		}
		return robots;	
	}
	
	public String plotRouteToPoint(Point p, int originX, int originY) throws Exception
	{
		// ENHANCEMENT: Cache computed routes and their inverses to avoid recalculating routes
		long startTime = System.currentTimeMillis();
		long endTime = 0;
		resetHopCount();
		
		String path = null;
//		System.out.println("Starting at: " + originX + ", " + originY);
//		System.out.println("Destination(in coordinates): " + p.toString());
		
		// convert to array index
		GridCell origin = m_Rows[originX - 1][originY - 1];
		int depth = 0;
		boolean found = false;
		
		// Test origin to see if its the destination
		if ((origin.x == p.x - 1) && (origin.y == p.y - 1))
		{
			throw new Exception();
//			return "X";	
		}
		
		// begin search
		GridCell currentCell = origin;
		currentCell.hopCount = 0;
		GridCell nextCell = null;
		Vector holder = new Vector();
		Vector newV = null;
		Vector oldV = new Vector();
//		System.out.println("Root (depth 0) at [" + Integer.toString(currentCell.x) + "," + Integer.toString(currentCell.y) + "]");
		oldV.add(currentCell);
		holder.add(oldV);
		while (!found)
		{
//			System.out.println(Integer.toString(oldV.size()) + " Elements in layer (depth " + Integer.toString(depth) + ")");
			if (oldV.size() == 0)
			{
//				System.out.println("Can't find destination?");
				return pickSafeDirection(originX, originY);
			}
			newV = new Vector();
			for (int i=0; i<oldV.size(); i++)
			{
				currentCell = (GridCell)oldV.get(i);
				if (currentCell.canGoNorth)
				{
//					System.out.println("Current Node can go North");
					nextCell = m_Rows[currentCell.x][currentCell.y + 1];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = SOUTH;
						nextCell.hopCount = depth;
						if ((nextCell.x == p.x - 1) && (nextCell.y == p.y - 1))
						{
//							System.out.println("Target Found!");
							newV.add(nextCell);	
							found = true;
							break;	
						}
//						System.out.println("Cell Added (depth " + Integer.toString(depth + 1) + ") at [" + Integer.toString(nextCell.x) + "," + Integer.toString(nextCell.y) + "]");
//						System.out.println("Previous vector size: " + newV.size());
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");
//						System.out.println("New vector size: " + newV.size());
					}
				}
				if (currentCell.canGoEast)
				{
//					System.out.println("Current Node can go East");
					nextCell = m_Rows[currentCell.x + 1][currentCell.y];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = WEST;
						nextCell.hopCount = depth;
						if ((nextCell.x == p.x - 1) && (nextCell.y == p.y - 1))
						{
//							System.out.println("Target Found!");
							newV.add(nextCell);	
							found = true;
							break;	
						}
//						System.out.println("Cell Added (depth " + Integer.toString(depth + 1) + ") at [" + Integer.toString(nextCell.x) + "," + Integer.toString(nextCell.y) + "]");
//						System.out.println("Previous vector size: " + newV.size());
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");	
//						System.out.println("New vector size: " + newV.size());
						
					}
				}
				if (currentCell.canGoSouth)
				{
//					System.out.println("Current Node can go South");
					nextCell = m_Rows[currentCell.x][currentCell.y - 1];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = NORTH;
						nextCell.hopCount = depth;
						if ((nextCell.x == p.x - 1) && (nextCell.y == p.y - 1))
						{
//							System.out.println("Target Found!");
							newV.add(nextCell);	
							found = true;
							break;	
						}
//						System.out.println("Cell Added (depth " + Integer.toString(depth + 1) + ") at [" + Integer.toString(nextCell.x) + "," + Integer.toString(nextCell.y) + "]");
//						System.out.println("Previous vector size: " + newV.size());
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");
//						System.out.println("New vector size: " + newV.size());
					}
				}
				if (currentCell.canGoWest)
				{
//					System.out.println("Current Node can go West");
					nextCell = m_Rows[currentCell.x - 1][currentCell.y];
					if (nextCell.hopCount == -1)
					{
						nextCell.parentDirection = EAST;
						nextCell.hopCount = depth;
						if ((nextCell.x == p.x - 1) && (nextCell.y == p.y - 1))
						{
//							System.out.println("Target Found!");
							newV.add(nextCell);	
							found = true;
							break;	
						}
//						System.out.println("Cell Added (depth " + Integer.toString(depth + 1) + ") at [" + Integer.toString(nextCell.x) + "," + Integer.toString(nextCell.y) + "]");
//						System.out.println("Previous vector size: " + newV.size());
						if (newV.add(nextCell) != true)
							System.out.println("Error Adding to Vector");
//						System.out.println("New vector size: " + newV.size());
					}
				}
			}
		
			oldV = newV;
			holder.add(oldV);
			depth++;
//			if (depth > 21) System.out.println("Lost finding path");
		}
		
		// now holder contains a map of all the nodes we've come across. Time to parse it
		// and determine the direction sequence to move from origin to destination (computed 
		// backwards of course).
		
		StringBuffer sequence = new StringBuffer();
		boolean nextHopFound = false;
		String temp = null;
		currentCell = (GridCell) ((Vector)holder.lastElement()).lastElement();
		for (int j=holder.size() - 2; j>=0; j--)
		{
			Vector layer = (Vector) holder.get(j);
			for (int k=0; k<layer.size(); k++)
			{
				nextCell = (GridCell) layer.get(k);
				if ((currentCell.parentDirection == NORTH) && (nextCell.x == currentCell.x) && (nextCell.y == currentCell.y + 1))
				{
					temp = "S ";
					nextHopFound = true;
					break;
				}
				if ((currentCell.parentDirection == EAST) && (nextCell.x == currentCell.x + 1) && (nextCell.y == currentCell.y))
				{
					temp = "W ";
					nextHopFound = true;
					break;
				} 
				if ((currentCell.parentDirection == SOUTH) && (nextCell.x == currentCell.x) && (nextCell.y + 1 == currentCell.y))
				{
					temp = "N ";
					nextHopFound = true;
					break;
				} 
				if ((currentCell.parentDirection == WEST) && (nextCell.x + 1 == currentCell.x) && (nextCell.y == currentCell.y))
				{
					temp = "E ";
					nextHopFound = true;
					break;
				}  	
			}
			if (nextHopFound)
			{
				// character to sequence
				sequence.insert(0, temp);
				currentCell = nextCell;
			}	
		}
		
		path = sequence.toString().trim();
//		System.out.println("Path to node calculated");
		endTime = System.currentTimeMillis();
//		System.out.println("Time spent pathfinding(ms): " + (int)(endTime - startTime));
//		System.out.println("-" + path);	
		return path;
	}
	
	public void resetHopCount()
	{
		for (int i=0; i<m_RowCountX; i++)
		{
			for (int j=0; j<m_ColCountY; j++)
			{
				GridCell gc = m_Rows[i][j];
				gc.hopCount = -1;	
			}	
		}
	}
	
	public String pickSafeDirection(int originX, int originY)
	{
		//remember args are coordinates
		GridCell gc = m_Rows[originX - 1][originY - 1];
		String direction = null;
		if (gc.canGoNorth)
			direction ="N";	 
		if (gc.canGoEast)
			direction ="E";
		if (gc.canGoSouth)
			direction ="S";
		if (gc.canGoWest)
			direction ="W";
		return direction;
	}
	
	public void addBotInterface(BotInterface bi)
	{
		m_OurBot = bi;	
	}
	
	public void updateRobotMovement(int id, char movement)
	{
//		System.out.println("---Update bot " + id + " going " + movement);
		boolean foundBot = false;
		Iterator iter = m_RobotCatalog.iterator();
		while (iter.hasNext())
		{
			Robot r = (Robot) iter.next();
			int rID = r.getID();
			if (rID == id)
			{
				foundBot = true;
				switch (movement)
				{
					case ('N'):
						r.setLocation(r.getLocationX(), r.getLocationY() + 1);
						break;
					case ('E'):
						r.setLocation(r.getLocationX() + 1, r.getLocationY());
						break;
					case ('S'):
						r.setLocation(r.getLocationX(), r.getLocationY() - 1);
						break;
					case ('W'):
						r.setLocation(r.getLocationX() - 1, r.getLocationY());
						break;
				}
				break;
			}		
		}
		if (foundBot == false)
		{
			// what to do here?
			System.out.println("Creating new bot?");
			m_RobotCatalog.add(new Robot(id, 1, 1));	
		}
	}
	
	public void updateRobotLocation(int id, int x, int y)
	{
		boolean foundBot = false;
		Iterator iter = m_RobotCatalog.iterator();
		while (iter.hasNext())
		{
			Robot r = (Robot) iter.next();
			int rID = r.getID();
			if (rID == id)
			{
				foundBot = true;
				r.setLocation(x,y);
				break;
			}		
		}
		if (foundBot == false)
		{
			System.out.println("Creating new bot[" + id + "]");
			m_RobotCatalog.add(new Robot(id, x, y));	
		}
	}
	
	public void updateRobotPackage(int id, boolean packageDrop, int packageID)
	{
		Iterator iter = m_RobotCatalog.iterator();
		while (iter.hasNext())
		{
			Robot r = (Robot) iter.next();
			int rID = r.getID();
			if (rID == id)
			{
				// so we got the bot, lets get the package
				Iterator packIterator = m_PackageCatalog.iterator();
				PackageObject po = null;
				while (packIterator.hasNext())
				{
					PackageObject p = (PackageObject)packIterator.next();
					if (p.getID() == packageID)
					{
						po = p;
						break;
					}		
				}
				if (po == null) return;
				if (packageDrop == true)
					r.removePackage(po);
				else
					r.addPackage(po);	
				
				break;
			}		
		}

	}
	
	public void updateCurrentRobots(Vector activeRobots)
	{
		// only have to worry about deletion here as if we get robot info from
		// server for robot that we don't have, its created and added to the catalog
//		System.out.println("active bots vector:\n" + activeRobots.toString());
//		System.out.println("robot catalog: \n" + m_RobotCatalog.toString());
		Iterator iter = m_RobotCatalog.iterator();
		while (iter.hasNext())
		{
			Robot r = (Robot) iter.next();
			String rID = Integer.toString(r.getID());
			if (!(activeRobots.contains(rID)))
			{
				System.out.println("Deleting robot from field: [" + rID + "]");
				iter.remove();
				break;
			}		
		}
	}

	public void updatePackageCatalog(int packageID, int x, int y, int owned)
	{
		boolean previouslyRecorded = false;
		for (int i=0; i<m_PackageCatalog.size(); i++)
		{
			PackageObject po = (PackageObject) m_PackageCatalog.get(i);
			if (po.getID() == packageID)
			{
				previouslyRecorded = true;
				po.setLocation(x,y);
				po.setState(owned);
			}	
		}
		if (previouslyRecorded == false)
		{
			PackageObject newP = new PackageObject(packageID);
			newP.setLocation(x, y);
			newP.setState(owned);
			m_PackageCatalog.add(newP);	
		}	
	}
	
	private int 			m_RowCountX;
	private int 			m_ColCountY;
	private GridCell [][] 	m_Rows;
	private Vector 			m_CurrentPackages;
	private Vector 			m_PackageCatalog;
	private Vector			m_RobotCatalog;	
	private Vector			m_HomeBases;
	private double			m_BotDensity;
	private double 			m_terrainDensity;
	private BotInterface	m_OurBot;
}