/* 
	OurBot.java
	Jason C. Judt
	This file was writen for the ICFP 2002 contest and may be used by anyone, anywhere
	without restriction. Enjoy. 
*/

package myprojects.jbotchallenge;

import java.awt.*;
import java.util.*;
import myprojects.jbotchallenge.GameGrid;
import myprojects.jbotchallenge.BotFunction;
import myprojects.jbotchallenge.SimpleBotFunction;
import myprojects.jbotchallenge.BotInterface;

public class OurBot extends Object implements BotInterface
{
	
	// seeking packages
	public static final Integer SEEKING_STATE = new Integer(0);
	public static final Integer DELIVERING_STATE = new Integer(1);
	public static final Integer EVADING_STATE = new Integer(2);
	public static final Integer HUNTING_STATE = new Integer(3);

	public OurBot(GameGrid gg)
	{
		m_StateTable = new Hashtable();
		m_GameGrid = gg;
		m_State = SEEKING_STATE;
		m_CurrentWeight = 0;
			
	}
	
	public void initialize(int id, int weightAllow, int money)
	{
		m_ID = id;
		m_MaxWeight = weightAllow;
		m_Money = money;
		m_Packages = new Vector();	
		loadStates();
		m_GameGrid.addBotInterface(this);
	}
	
	protected void loadStates()
	{
		m_StateTable.put(SEEKING_STATE, new BPSeekingBotFunction(m_GameGrid, this));
		m_StateTable.put(DELIVERING_STATE, new SimpleDeliveryBotFunction(m_GameGrid, this)); 
		m_StateTable.put(EVADING_STATE, new SimpleEvadingBotFunction(m_GameGrid, this)); 
		m_StateTable.put(HUNTING_STATE, new SimpleBotFunction(m_GameGrid, this)); 	
	}
	
	public String getNextAction()
	{
		Vector closeRobots = m_GameGrid.getRobotsWithinTwoStepsOfPoint(m_LocationX, m_LocationY);
		if (closeRobots.size() > 0)
		{
			changeState(EVADING_STATE);
		}
		else
		{
			if (m_State == EVADING_STATE)
			{
				// got away...back to normal
				if (m_Packages.size() > 0) changeState(DELIVERING_STATE);
				else changeState(SEEKING_STATE);
			}
		}
		BotFunction bf = (BotFunction) m_StateTable.get(m_State);
		System.out.println("Current State is " + m_State.toString());
		if (bf == null)
			System.out.println("No bot action");
		String nextMove = bf.calcNextMove();
		System.out.println("Next Move!!! " + nextMove);
		StringTokenizer actionFinder = new StringTokenizer(nextMove, " ");
		String bid = actionFinder.nextToken();		// get rid of bid
		System.out.println("Bid - " + bid);
		m_Money -= Integer.parseInt(bid); 
		String action = actionFinder.nextToken();
		if (action.equals("Move"))
		{
			String direction = actionFinder.nextToken();
			char d = direction.charAt(0);
			switch (d)
			{
				case ('N'):
					m_NextPoint = new Point(m_LocationX, m_LocationY + 1);
					break;
				case ('S'):
					m_NextPoint = new Point(m_LocationX, m_LocationY - 1);
					break; 
				case ('E'):
					m_NextPoint = new Point(m_LocationX + 1, m_LocationY);
					break;
				case ('W'):
					m_NextPoint = new Point(m_LocationX - 1, m_LocationY);
					break;
			}
		}
		else
		{
			m_NextPoint = null;
		}
		if (action.equals("Drop"))
		{
			m_DropList = "";
			while (actionFinder.hasMoreTokens())
			{
				m_DropList += actionFinder.nextToken();
				m_DropList += " ";
			}		
			m_DropList = m_DropList.trim();
		}
		else
		{
			m_DropList = null;
		}				
		return nextMove;		
	}
	
	public void changeState(Integer nextState)
	{
		BotFunction bf = (BotFunction) m_StateTable.get(m_State);
//		new Exception().printStackTrace();
		System.out.println("Leaving State " + m_State.toString());
		bf.cleanUp();
		m_State = nextState;
	}
	
	public void updatePosition(char direction)
	{
		switch (direction)
		{
			case ('N'):
				m_LocationY++;
				break;
			case ('E'):
				m_LocationX++;
				break;
			case ('W'):
				m_LocationX--;
				break;
			case ('S'):
				m_LocationY--;
				break;
		}
		System.out.println("OurBot New Location (" + Integer.toString(m_LocationX) + "," + Integer.toString(m_LocationY) + ")");
		int x = m_GameGrid.getCellTypeAt(m_LocationX, m_LocationY);
		switch (x)
		{
			case (GameGrid.HOME_BASE):
				System.out.println("OurBot at a base");
				break;
			case (GameGrid.WALL):
				System.out.println("Error, OurBot is in a wall?");
				break;
			case (GameGrid.WATER):
				System.out.println("Error, OurBot is swimming!");
				break;
			case (GameGrid.OPEN_SPACE):
				System.out.println("OurBot is in the open");
		}
		
		for (int i=0; i<m_Packages.size(); i++)
		{
			PackageObject po = (PackageObject) m_Packages.get(i);
			po.setLocation(m_LocationX, m_LocationY);
		}
		
		if (m_NextPoint != null)
		{
			Point currentPoint = new Point(m_LocationX, m_LocationY);
			if (!(currentPoint.equals(m_NextPoint)))
			{
				System.out.println("BUMPED!!!");
				// reset current bot function so it will recover
				BotFunction bf = (BotFunction) m_StateTable.get(m_State);
				bf.cleanUp();
			}
		}
	}
	
	public void updateLocation(int x, int y)
	{
		m_LocationX = x;
		m_LocationY = y;
		System.out.println("OurBot New Location (" + Integer.toString(x) + "," + Integer.toString(y) + ")");
		int z = m_GameGrid.getCellTypeAt(m_LocationX, m_LocationY);
		switch (z)
		{
			case (GameGrid.HOME_BASE):
				System.out.println("OurBot at a base");
				break;
			case (GameGrid.WALL):
				System.out.println("Error, OurBot is in a wall?");
				break;
			case (GameGrid.WATER):
				System.out.println("Error, OurBot is swimming!");
				break;
			case (GameGrid.OPEN_SPACE):
				System.out.println("OurBot is in the open");
				break;
		}
		
		for (int i=0; i<m_Packages.size(); i++)
		{
			PackageObject po = (PackageObject) m_Packages.get(i);
			po.setLocation(m_LocationX, m_LocationY);
		}
		
		if (m_NextPoint != null)
		{
			Point currentPoint = new Point(m_LocationX, m_LocationY);
			if (!(currentPoint.equals(m_NextPoint)))
			{
				System.out.println("BUMPED<Location mismatch>");
				// reset current bot function so it will recover
				BotFunction bf = (BotFunction) m_StateTable.get(m_State);
				bf.cleanUp();
			}
		}
		
	}
	
	public void updatePackages(boolean dropped, int packageID)
	{
		if (dropped)
		{
//			m_Packages.remove(new Integer(packageID));	
			for (int i=0; i<m_Packages.size(); i++)
			{
				PackageObject po = (PackageObject) m_Packages.get(i);
				if (po.getID() == packageID)
				{
					m_CurrentWeight -= po.getWeight();
//					po.setLocation(m_LocationX, m_LocationY);
					po.setState(PackageObject.DROPPED_STATE);
					// see if this drop was planned
					if (m_DropList != null)
					{
						StringTokenizer st = new StringTokenizer(m_DropList, " ");
						boolean planned = false;
						while (st.hasMoreTokens())
						{
							if (Integer.parseInt(st.nextToken()) == packageID)
							{
								planned = true;
								break;
							}    
						}
						if (planned == false)
						{
//							System.out.println("BUMPED<Lost Package>");
						}
						else
						{
							Iterator iter = m_GameGrid.getPackageCatalog().iterator();
							while (iter.hasNext())
							{
								PackageObject temp = (PackageObject) iter.next();
								if (temp.getID() == po.getID())
								{
									iter.remove();
									break;
								}
							}
						}
					}
					m_Packages.remove(po);
					break;
				}
			}
		}
		else
		{
//			m_Packages.add(new Integer(packageID));	
			
			for (int i=0; i<m_GameGrid.getCurrentPackage().size(); i++)
			{
				PackageObject po = (PackageObject) m_GameGrid.getCurrentPackage().get(i);
				if (po.getID() == packageID)
				{
					m_CurrentWeight += po.getWeight();
					po.setLocation(m_LocationX, m_LocationY);
					po.setState(PackageObject.OWNED_STATE);
					m_Packages.add(po);
					break;
				}
			}
					
		}
		if (m_Packages.size() > 0)
		{
			// switch states to Deliver?
			changeState(DELIVERING_STATE);
//			System.out.println("Carrying weight: " + m_CurrentWeight);	
		}
		if (m_Packages.size() == 0)
		{
			// switch to seek?
			changeState(SEEKING_STATE);
		}
	}
	
	public int getMoney()
	{
		return m_Money;
	}
	
	public int getMaxWeight()
	{
		return m_MaxWeight;
	}
	
	public int getCurrentWeight()
	{
		return m_CurrentWeight;
	}	
	
	public int getID()
	{
		return m_ID;	
	}
	
	public int getLocationX()
	{
		return m_LocationX;
	}
	
	public int getLocationY()
	{
		return m_LocationY;	
	}
	
	public Vector getPackages()
	{
		return m_Packages;	
	}
	
	public void setCurrentPath(String path)
	{
		m_CurrentPath = path;
	}
	
	public String getCurrentPath()
	{
		return m_CurrentPath;
	}
	
	private	int 		m_Money;
	private int 		m_MaxWeight;
	private int			m_CurrentWeight;
	private int 		m_ID;
	private int			m_LocationX;
	private int 		m_LocationY;		
	private GameGrid	m_GameGrid;
	protected Integer	m_State;
	protected Hashtable	m_StateTable;
	protected Vector	m_Packages;
	private Point		m_NextPoint;
	private String		m_DropList;
	
	
	// not in use
	private String		m_CurrentPath;
	private Point		m_CurrentDestination;
}