package myprojects.jbotchallenge;

import java.awt.Point;
import java.util.*;
import myprojects.jbotchallenge.BotFunction;

public class SimpleDeliveryBotFunction extends BotFunction
{
	public SimpleDeliveryBotFunction(GameGrid gg, BotInterface bot)
	{
		super(gg, bot);	
//		m_NoPackagesCollection = new Vector();	
		m_BidDefault = "1";
		m_PathToDestination = null;
		m_RecalcDestination = false;
	}
	
	public String calcNextMove()
	{
		
		System.out.println("OurBot<Delivery> choosing action");
		m_NextActionString = null;
		// first, let's see if we are on a delivery destination
		Vector packCatalog = m_gg.getPackageCatalog();
		StringBuffer sb = new StringBuffer();
		for (int m=0; m<packCatalog.size(); m++)
		{
			PackageObject po = (PackageObject) packCatalog.elementAt(m);
			if (m_bot.getPackages().contains(po))
			{
				if (po.getDestinationX() == m_bot.getLocationX())
				{
					if (po.getDestinationY() == m_bot.getLocationY())
					{
						sb.insert(0, Integer.toString(po.getID()) + " ");
						
					}
				}	
			}
		}
		if (sb.length() > 0)
		{
			String dropList = sb.toString().trim();
			m_NextActionString = determineBid() + " Drop " + dropList;
			m_RecalcDestination = true;	
			return m_NextActionString;
		}
		
		// second, find out how many packages we have
		if (m_gg.getCurrentPackage().size() > 0)
		{
			System.out.println("OurBot<Delivery> has stumbled across some packages"); // we have packages
			// if we can, let's grab all of the lightest ones
			String packageList = choosePackages();
			if (packageList != null)
			{
				m_NextActionString = determineBid() + " Pick " + packageList;
				m_RecalcDestination = true;	
				return m_NextActionString;
			}
			else 
			{
				// couldn't add any packages	
			}
		}

		
		if ((m_PathToDestination == null) || (m_RecalcDestination == true)) 
			m_PathToDestination = calculatePath();
		
/*		if (m_PathToDestination.equals("X")
		{
				
		}	
*/
		m_NextActionString = determineBid() + " Move " + popNextPathSegment();
			
		return m_NextActionString;
	}
	
	protected String choosePackages()
	{
		String packages = null;
		StringBuffer packBuffer = new StringBuffer();
		Vector packagesToGet = m_gg.getCurrentPackage();
		Vector localVect = new Vector();
		for (int i=0; i<packagesToGet.size(); i++)
		{
			localVect.add(packagesToGet.get(i));	
		}
		// for now, get the least expensive packages and pick them up.
		int packageWeight = Integer.MAX_VALUE;
		int weightAllowance = m_bot.getMaxWeight() - m_bot.getCurrentWeight();
		while (localVect.size() > 0) 	// have to break out manually if we hit weight limit
		{
			int j=0;
			int index = 0;
			for (j=0; j<localVect.size(); j++)
			{
				PackageObject po = (PackageObject)localVect.get(j);
				if (po.getWeight() < packageWeight)
				{
					packageWeight = po.getWeight();
					index = j;	
				}
			}
			// now we have the index of the lightest...lets add that to our list if we can
			PackageObject wantedPackage = (PackageObject)localVect.get(index);
			if (wantedPackage.getWeight() <=  weightAllowance)
			{
				packBuffer.append(wantedPackage.getID());
				packBuffer.append(' ');
				localVect.remove(index);
				weightAllowance -= wantedPackage.getWeight();
			}
			else		// manual abort
			{
				break;
			}
				
		} 
		if (packBuffer.length() > 0)
			packages = packBuffer.toString().trim();
		return packages;
	}
	
	protected String calculatePath()
	{
		m_RecalcDestination = false;
		String result = null;
		int closestDeliveryIndex = -1;
		int closestDeliveryDist = Integer.MAX_VALUE;
		int i = 0;
		System.out.println("OurBot<Delivery> calculating path to Destination");
		
		// look through catalog to find close packages and keep the one with the closest destination
		// ENHANCEMENT: go ahead and calculate a path for all packages being held. Then do some fancy
		// 				determining if its better to go to a group of destinations rather than just the
		// 				closest.
		Vector packCatalog = m_gg.getPackageCatalog();
		for (i=0; i<packCatalog.size(); i++)
		{
			PackageObject po = (PackageObject) packCatalog.elementAt(i);
			if (m_bot.getPackages().contains(po))
			{
				// po is a package we have picked up
				int dist = PackageObject.getShortestDistance(po, m_bot.getLocationX(), m_bot.getLocationY()); 
				if (dist < closestDeliveryDist)
				{
					// ENHANCEMENT: maybe we should make sure the destination isn't water or a wall or something
					closestDeliveryIndex = i;
					closestDeliveryDist = dist;
						
				}
			}
		}
		
		
		if (closestDeliveryIndex == -1)
		{
			// hmmm...nothing found. Let's move randomly
			System.out.println("OurBot<Delivery> failed to find any destination");
//			result = "N";	// don't pick N...check N, if valid, use, if not, choose E, then S, then W
			result = m_gg.pickSafeDirection(m_bot.getLocationX(), m_bot.getLocationY());
			return result;	
		}
		
		Point newDestination = null;
		
		System.out.println("OurBot<Delivery> found destination");
		PackageObject poNew = (PackageObject) packCatalog.elementAt(closestDeliveryIndex);	
		newDestination = new Point(poNew.getDestinationX(), poNew.getDestinationY());	
		
		result = computeRouteToPoint(newDestination);
		
		return result;
	}
	
	protected String determineBid()
	{
		return m_BidDefault;
	}
	
	private String popNextPathSegment()
	{
		String segment = null;
//		System.out.println("OurBot<Delivery> old path is: " + m_PathToDestination);
		if (m_PathToDestination.length() <= 2)
		{
			segment = m_PathToDestination;
			m_PathToDestination = null;
			return segment;
		}	
		segment = m_PathToDestination.substring(0, 1);
		System.out.println("OurBot<Delivery> will move " + segment);
		m_PathToDestination = m_PathToDestination.substring(2);
		m_bot.setCurrentPath(m_PathToDestination);
//		System.out.println("OurBot<Delivery> future path is: " + m_PathToDestination);
		return segment;	
	}
	
	protected void cleanUp()
	{
		// reset so this state can be run again in a new context
		m_PathToDestination = null;
		m_NextActionString = null;
		m_RecalcDestination = false;
	}
		
	private String computeRouteToPoint(Point p)
	{
		String route = null;
		try
		{
			route = m_gg.plotRouteToPoint(p, m_bot.getLocationX(), m_bot.getLocationY());
		}
		catch (Exception e)
		{
			System.out.println("Error calculating route: " + e.getMessage());
			e.printStackTrace();
			route = m_gg.pickSafeDirection(m_bot.getLocationX(), m_bot.getLocationY());
		} 
			
		return route;
	}
	
//	private Vector m_NoPackagesCollection;	// vector of points where we have been that have no packages
	private String 	m_PathToDestination;
	private String 	m_NextActionString;
	private String 	m_BidDefault;
	private boolean m_RecalcDestination;
}