/* 
	SimpleSeekingBotFunction.java
	Jason C. Judt
	This file was writen for the ICFP 2002 contest and may be used by anyone, anywhere
	without restriction. Enjoy. 
*/

package myprojects.jbotchallenge;

import java.awt.Point;
import java.util.*;
import myprojects.jbotchallenge.BotFunction;

public class SimpleSeekingBotFunction extends BotFunction
{
	public SimpleSeekingBotFunction(GameGrid gg, BotInterface bot)
	{
		super(gg, bot);	
		m_NoPackagesCollection = new Vector();	
		m_BidDefault = "1";
		m_PathToPackages = null;
	}
	
	public String calcNextMove()
	{
		// first, find if we are on any packages
		System.out.println("OurBot<Seeking> choosing action");
		m_NextActionString = null;
		if (m_gg.getCurrentPackage().size() > 0)
		{
//			System.out.println("OurBot<Seeking> has found packages"); // we have packages
			// if we can, let's grab all of the lightest ones
			String packageList = choosePackages();
			if (packageList != null)
			{
				m_NextActionString = determineBid() + " Pick " + packageList;	
				return m_NextActionString;
			}
			else 
			{
				// couldn't add any packages	
			}
		}
		else
		{
			m_NoPackagesCollection.add(new Point(m_bot.getLocationX(), m_bot.getLocationY()));
		}
//		System.out.println("No packages for Ourbot<Seeking> here");
		// let's add this point to the "no packages" vector so we won't come back here
		
		
		if (m_PathToPackages == null)
			m_PathToPackages = calculatePath();
			
		m_NextActionString = determineBid() + " Move " + popNextPathSegment();
			
		return m_NextActionString;
	}
	
	protected String choosePackages()
	{
		String packages = null;
		StringBuffer packBuffer = new StringBuffer();
		Vector packagesToGet = m_gg.getCurrentPackage();
		Vector localVect = new Vector();
		// make a local copy of vector that we can modify
		for (int i=0; i<packagesToGet.size(); i++)
		{
			localVect.add(packagesToGet.get(i));	
		}
		// for now, get the least expesive packages and pick them up.
		// ENHANCEMENT: check to make sure packages have a "good" drop off point. 
		// ENHANCEMENT: maybe do a blend of gathering smallest and closest destinations
		int packageWeight = Integer.MAX_VALUE;
		int weightAllowance = m_bot.getMaxWeight() - m_bot.getCurrentWeight();
		while (localVect.size() > 0) 	// have to break out manually if we hit weight limit
		{
			int j=0;
			int index = 0;
			for (j=0; j<localVect.size(); j++)
			{
				PackageObject po = (PackageObject)localVect.get(j);
				if (po.getWeight() < packageWeight)
				{
					packageWeight = po.getWeight();
					index = j;	
				}
			}
			// now we have the index of the lightest...lets add that to our list if we can
			PackageObject wantedPackage = (PackageObject)localVect.get(index);
			if (wantedPackage.getWeight() <=  weightAllowance)
			{
				packBuffer.append(wantedPackage.getID());
				packBuffer.append(' ');
				localVect.remove(index);
				weightAllowance -= wantedPackage.getWeight();
			}
			else		// manual abort
			{
				break;
			}
				
		} 
		if (packBuffer.length() > 0)
			packages = packBuffer.toString().trim();
		return packages;
	}
	
	protected String calculatePath()
	{
		String result = null;
		int closestPackageIndex = -1;
		int closestPackageDist = Integer.MAX_VALUE;
		int i = 0;
//		System.out.println("OurBot<Seeking> calculating path to packages");
		
		// look through catalog to find close packages and keep closest one
		Vector packCatalog = m_gg.getPackageCatalog();
		for (i=0; i<packCatalog.size(); i++)
		{
			PackageObject po = (PackageObject) packCatalog.elementAt(i);
		
			int dist = PackageObject.getShortestDistance(po, m_bot.getLocationX(), m_bot.getLocationY()); 
			// lets try calculating the precise distance
	//		StringTokenizer st1 = new StringTokenizer(m_gg.plotRouteToPoint(p, m_bot.getLocationX(), m_bot.getLocationY())
//	int dist = 
			if (dist < closestPackageDist)
			{
				// ENHANCEMENT: maybe we should only do this when we know the package is definately there 
				// 				in otherwords, maybe we should check the package state.
				boolean pointReject = false;
				Point pp = new Point(po.getLocationX(), po.getLocationY());
				for (int j=0; j<m_NoPackagesCollection.size(); j++)
				{
					Point tempPackagePoint = (Point) m_NoPackagesCollection.get(j);
					if (tempPackagePoint.x == pp.x)
					{
						if (tempPackagePoint.y == pp.y)
						{
							pointReject = true;
						}
					}
				}
				if (pointReject == false)
				{
					closestPackageIndex = i;
					closestPackageDist = dist;
				}	
			}
		}
		
		// now if closestPackageID != -1 then we have the closest package known
		// let's see if a home base is closer and if it has packages or not	
		Vector bases = m_gg.getHomeBases();
		int closestBaseIndex = -1;
		int closestBaseDistance = Integer.MAX_VALUE;
		for (i=0; i<bases.size(); i++)
		{
			Point pb = (Point) bases.get(i);
			int baseDistance = computeDistance(pb, new Point(m_bot.getLocationX(), m_bot.getLocationY()));
				
			if (baseDistance < closestBaseDistance)
			{
				boolean baseReject = false;
				for (int k=0; k<m_NoPackagesCollection.size(); k++)
				{
					Point tempBasePoint = (Point) m_NoPackagesCollection.get(k);
					if (tempBasePoint.x == pb.x)
					{
						if (tempBasePoint.y == pb.y)
						{
							baseReject = true;
						}
					}
				}
				if (baseReject == false)
				{
					closestBaseIndex = i;
					closestBaseDistance = baseDistance;
				}	
			} 	
		}
		
		if ((closestPackageIndex == -1) && (closestBaseIndex == -1))
		{
			// hmmm...nothing found. Let's move randomly
//			System.out.println("OurBot<Seeking> failed to aquire any target");
			result = m_gg.pickSafeDirection(m_bot.getLocationX(), m_bot.getLocationY());	// don't pick N...check N, if valid, use, if not, choose E, then S, then W
			return result;	
		}
		
		Point newDestination = null;
		
		if (closestBaseDistance <= closestPackageDist)	// let's favor bases as they tend to have more packages
		{
//			System.out.println("OurBot<Seeking> found base");
			newDestination = (Point) bases.get(closestBaseIndex);		
		}
		else
		{
//			System.out.println("OurBot<Seeking> found package");
			newDestination = new Point(((PackageObject)packCatalog.get(closestPackageIndex)).getLocationX(), ((PackageObject)packCatalog.get(closestPackageIndex)).getLocationX());;	
		}
		result = computeRouteToPoint(newDestination);
		
		return result;
	}
	
	protected String determineBid()
	{
		return m_BidDefault;
	}
	
	private String popNextPathSegment()
	{
		String segment = null;
//		System.out.println("OurBot<Seeking> old path is: " + m_PathToPackages);
		if (m_PathToPackages.length() <= 2)
		{
			segment = m_PathToPackages;
			m_PathToPackages = null;
			return segment;
		}	
		segment = m_PathToPackages.substring(0, 1);
//		System.out.println("OurBot<Seeking> will move " + segment);
		m_PathToPackages = m_PathToPackages.substring(2);
//		System.out.println("OurBot<Seeking> future path is: " + m_PathToPackages);
		m_bot.setCurrentPath(m_PathToPackages);
		if (segment.equals("X"))
		{
			int jason = 0;
			jason++;
		}
		return segment;	
	}
	
	protected void cleanUp()
	{
		// reset so this state can be run again in a new context
		m_PathToPackages = null;
		m_NextActionString = null;
	}
		
	protected String computeRouteToPoint(Point p)
	{
		String route = null;
		try
		{
			route = m_gg.plotRouteToPoint(p, m_bot.getLocationX(), m_bot.getLocationY());
		}
		catch (Exception e)
		{
//			System.out.println("Error calculating route: " + e.getMessage());
//			e.printStackTrace();
			route = m_gg.pickSafeDirection(m_bot.getLocationX(), m_bot.getLocationY());
		} 
			
		return route;
	}
	
	protected Vector m_NoPackagesCollection;	// vector of points where we have been that have no packages
	private String m_PathToPackages;
	private String m_NextActionString;
	private String m_BidDefault;
}