/* 
	SocketClient.java
	Jason C. Judt
	This file was writen for the ICFP 2002 contest and may be used by anyone, anywhere
	without restriction. Enjoy. 
*/

package myprojects.jbotchallenge;

import java.net.*;
import java.io.*;
import java.util.*;
import myprojects.jbotchallenge.GameGrid;
//import myprojects.jbotchallenge.Robot;

public class SocketClient extends Object
{

	public static final String INIT_STRING = "Player";
	
	public static final int INIT_STATE 			= 0;
	public static final int LIST_PACKAGE_STATE 	= 1;
	public static final int CLIENT_ACTION_STATE	= 2;
	public static final int SERVER_UPDATE_STATE	= 3;
	
	public SocketClient()
	{
		
	}	
	
	public int start(String serverHost, short serverPort, GameGrid gg, OurBot bot)
	{	
		StringTokenizer st = null;
		long startMark = 0;
		long endMark = 0;
		int numberOfRounds = 0;
		int cpuTimeConsumed = 0;
		int averageTimePerRound = 0;
		
		try
		{
			// init socket
			Socket s = new Socket(InetAddress.getByName(serverHost), serverPort);
			m_LNReader = new LineNumberReader(new BufferedReader(new InputStreamReader(s.getInputStream())));
			m_PrintWriter = new PrintWriter(new OutputStreamWriter(s.getOutputStream()));
			
			// write hello
			m_PrintWriter.println(INIT_STRING);
			m_PrintWriter.flush();
			
			// read board size
			String boardSize = m_LNReader.readLine();
			st = new StringTokenizer(boardSize, " "); 
			if (st.countTokens() != 2)
			{
				System.out.println("Error receiving board dimensions");
			}
			String xSize = st.nextToken();
			String ySize = st.nextToken();
			
			gg.initialize(Integer.parseInt(xSize), Integer.parseInt(ySize));
			
			// get board data
//			System.out.println("Board info: ");
			for (int i=1; i<=gg.getColCount(); i++)
			{
				String boardInfo = m_LNReader.readLine();
//				System.out.println("-" + boardInfo);
				gg.readRow(boardInfo, i);
			}
			
			// get id, weight allowance, and money
			st = null;
			String botInfo = m_LNReader.readLine();
//			System.out.println("Bot info: ");
//			System.out.println("-" + botInfo);
			st = new StringTokenizer(botInfo, " ");
			if (st.countTokens() != 3)
			{
//				System.out.println("Error receiving local bot info");
			}
			String id = st.nextToken();
			if (id.charAt(0) == '#')
				id = id.substring(1);
			String weight = st.nextToken();
			String money = st.nextToken();
			
			bot.initialize(Integer.parseInt(id), Integer.parseInt(weight), Integer.parseInt(money));
			
			// got most initialization. now lets start with turn #1 in loop below
			// must set to correct state initially. Also, signal grid to assign default
			// attraction values.
			//gg.updateGradients();
			gg.preprocessGrid();
			m_State = SERVER_UPDATE_STATE;
				
		}
		catch (Exception ioe)
		{
			System.out.println("Error initializing connection to server");	
			System.out.println(ioe.getMessage());
//			ioe.printStackTrace();
		}
		
		try 
		{
			while (true)
			{
				st = null;
				switch (m_State)
				{	
					case (SERVER_UPDATE_STATE):
					{
//						System.out.println("Update Info From Server:");
						String s0 = m_LNReader.readLine(); 
//						System.out.println(s0);
						st = new StringTokenizer(s0, "#");
						Vector thisRoundBots = new Vector();
						while (st.hasMoreTokens())
						{
							String s1 = st.nextToken();
//							System.out.println("-Bot Update:");
//							System.out.println("-" + s1);
							StringTokenizer stTemp = new StringTokenizer(s1, " ");
							int count = stTemp.countTokens();
							String idString = stTemp.nextToken();
							thisRoundBots.add(idString);
							int id = Integer.parseInt(idString);
							while (stTemp.hasMoreTokens())
							{
								String specificUpdate = stTemp.nextToken();
								specificUpdate = specificUpdate.toUpperCase();
								char ch = specificUpdate.charAt(0);
								switch (ch)
								{
									case ('N'):
									case ('S'):
									case ('E'):
									case ('W'):
//										System.out.println("--Update Bot (" + id + ") Movement - " + specificUpdate);
										if (id != bot.getID())
											gg.updateRobotMovement(id, ch);
										else 
											bot.updatePosition(ch);
										break;
									case ('X'):
										String xString = stTemp.nextToken(); 
										int x = Integer.parseInt(xString);
										stTemp.nextToken();// 'Y'
										String yString = stTemp.nextToken(); 	
										int y = Integer.parseInt(yString);
//										System.out.println("--Update Bot (" + id + ") Location - (" + xString + ", " + yString + ")");
										
										// change to array indexes from coordinates for game grid
										// but leave outbot alone
										if (id != bot.getID())
											gg.updateRobotLocation(id, x-1, y-1);
										else
											bot.updateLocation(x, y);
										break;
									case ('P'):
									case ('D'):
										boolean packageDrop = false;
										if (ch == 'D') packageDrop = true;
										String packageIDString = stTemp.nextToken();
										int packageID = Integer.parseInt(packageIDString);
//										System.out.println("--Update Bot (" + id + ") Package - ID: " + packageIDString + " dropped: " + packageDrop);
										if (id != bot.getID())
											gg.updateRobotPackage(id, packageDrop, packageID);
										else 
											bot.updatePackages(packageDrop, packageID);
										break;
								} 
							}
							
						}
						gg.updateCurrentRobots(thisRoundBots);	
						m_State = LIST_PACKAGE_STATE;
						System.out.println("\n------------ end round ------------");
						System.out.println("------------ time (ms) " + (endMark - startMark) + " ------------");
						System.out.println("------------ aver (ms) " + averageTimePerRound + " ------------");
						
						break;	
					}
					
					case (CLIENT_ACTION_STATE):
					{
//						System.out.println("Sending command to server: ");
						String action = bot.getNextAction();
//						System.out.println("-" + action);
						m_PrintWriter.println(action);
						m_PrintWriter.flush();
						endMark = System.currentTimeMillis();
						cpuTimeConsumed += (int)(endMark - startMark);
						averageTimePerRound = cpuTimeConsumed / numberOfRounds;
						m_State = SERVER_UPDATE_STATE;
						break;	
					}
					
					case (LIST_PACKAGE_STATE):
					{
						numberOfRounds++;
						System.out.println("\n############ round " + numberOfRounds + " ############\n");
						Vector packages = new Vector();
//						System.out.println("Package Info Received From Server:");
						String sp = m_LNReader.readLine(); 
						if (!(sp == null))
						{
							if (sp.startsWith("Robot"))
							{
								while (sp != null)
								{
									System.out.println(sp);
									sp = m_LNReader.readLine();
								}	
							}
						}
//						System.out.println("-" + sp);
						if (sp == null)
						{
//							System.out.println("Total CPU time (ms): " + cpuTimeConsumed); 
//							System.out.println("Program Terminated");
							return 0;
						}
						st = new StringTokenizer(sp, " ");
						while (st.hasMoreTokens())
						{
							String pkID = st.nextToken();
							String destX = st.nextToken();
							String destY = st.nextToken();
							String weight = st.nextToken();
							
							PackageObject po = new PackageObject(Integer.parseInt(pkID), Integer.parseInt(destX), Integer.parseInt(destY), Integer.parseInt(weight));
							packages.add(po);	
						}
						gg.setCurrentPackages(packages);
						startMark = System.currentTimeMillis();
						m_State = CLIENT_ACTION_STATE;					
						break;	
					}
				}
				Thread.sleep(10);
			}	
		}
		catch (Exception e)
		{
		
			System.out.println("Exception in server communication:");
			System.out.println(e.getMessage());	
//			e.printStackTrace();
			try
			{
				String sp = m_LNReader.readLine(); 
				if (!(sp == null))
				{
					if (sp.startsWith("Robot"))
					{
						while (sp != null)
						{
							System.out.println(sp);
							sp = m_LNReader.readLine();
						}	
					}
				}
			}
			catch (Exception ignored)
			{
			}
		}
		return 0;	
	}
	
	private LineNumberReader 	m_LNReader;
	private PrintWriter			m_PrintWriter;
	private int 				m_State;
}
