/* 
	BPSeekingBotFunction.java
	Jason C. Judt
	This file was writen for the ICFP 2002 contest and may be used by anyone, anywhere
	without restriction. Enjoy. 
*/

package myprojects.jbotchallenge;

import java.awt.Point;
import java.util.*;
import myprojects.jbotchallenge.BotFunction;
import myprojects.jbotchallenge.SimpleSeekingBotFunction;

public class BPSeekingBotFunction extends SimpleSeekingBotFunction
{
	
	public BPSeekingBotFunction(GameGrid gg, BotInterface bi)
	{
		super(gg, bi);	
	}
	
	protected String choosePackages()
	{
		int i = 0;
		String packages = null;
		StringBuffer packBuffer = new StringBuffer();
		Vector packagesToGet = m_gg.getCurrentPackage();
		Vector localVect = new Vector();		// holds all packages initially
		Vector sortedGroupVect = null;
		Vector groupVect = null;
		Vector packageVect = null;
		// make a local copy of vector that we can modify
		for (i=0; i<packagesToGet.size(); i++)
		{
			localVect.add(packagesToGet.get(i));	
		}
		
		// remove any packages that are to be delivered to a all square or water
		// also remove any package that has a weight bigger than the max bot allowance
		Iterator localIterator = localVect.iterator();
		while (localIterator.hasNext())
		{
			PackageObject po = (PackageObject) localIterator.next();
			if (po.getWeight() > m_bot.getMaxWeight()) 
			{
				po.setState(PackageObject.FUTILE_STATE);
				localIterator.remove();
				continue;
			}	
			int poX = po.getDestinationX();
			int poY = po.getDestinationY();	
			int tType = m_gg.getCellTypeAt(poX, poY);
			switch (tType)
			{
				case (GameGrid.WATER):
				case (GameGrid.WALL):
					po.setState(PackageObject.FUTILE_STATE);
					localIterator.remove();
					break;
				default:
					break;	
			}
		}
		
		// the following will demolish localVector
		// now group the sorted vector
		groupVect = groupVector(localVect);
		
		// sort the packages and groups
		sortedGroupVect = SortVectorByWeigthDistance(groupVect);
		
		// maybe I should sort in resort order by which groups are closer together
		
		 
		// unusual concept...once i have everything in the order I wished I could
		// grab it regardless of weight limits, there is no point in figuring out
		// which ones I could hold and only asking for those...the server will do 
		// all the work for me if I just know the order of preference I would like
		// the packages.
		
		// compute String
		for (i=0; i<sortedGroupVect.size(); i++)
		{
			Group gp = (Group) sortedGroupVect.get(i);
			Vector vTemp = gp.getPackages();
			for (int j=0; j<vTemp.size(); j++)
			{
				PackageObject pTemp = (PackageObject) vTemp.get(j);
				packBuffer.append(pTemp.getID());
				packBuffer.append(' ');
			}
		}
		
		if (packBuffer.length() > 0)
			packages = packBuffer.toString().trim();
		return packages;
	}
	
	protected Vector SortVectorByWeigthDistance(Vector groupVector)	
	{
		Vector result = new Vector();
		double maxWDValue = -1;
		int loopCount = 0;
		int index = 0;
		while (groupVector.size() > 0)
		{
			index = 0;
			maxWDValue = -1;
			for (int i=0; i<groupVector.size(); i++)
			{
				Group g = (Group) groupVector.get(i);
				double tempWDValue = g.getWDValue();
				if (tempWDValue >= maxWDValue)
				{
					index = i;
					maxWDValue = tempWDValue;	
				}
			}
			
			result.add(groupVector.remove(index));
		}
		
		
		return result;
		
	}
	
	protected Vector groupVector(Vector localVector)
	{
		Vector result = new Vector();
		int index = 0;
		int weightLimit = m_bot.getMaxWeight() - m_bot.getCurrentWeight();
		while (localVector.size() > 0)
		{
			index = 0;
			Vector collection = new Vector();
			Iterator collectionIterator = localVector.iterator();
			if (!collectionIterator.hasNext()) break;
			PackageObject firstItem = (PackageObject)collectionIterator.next(); 
			int targetX = firstItem.getDestinationX();
			int targetY = firstItem.getDestinationY();
			collection.add(firstItem);
			collectionIterator.remove();
			
			// find other package in vector with same destination
			while (collectionIterator.hasNext())
			{
				PackageObject po = (PackageObject) collectionIterator.next();
				int tempTargetX = po.getDestinationX();
				int tempTargetY = po.getDestinationY();
				if (tempTargetX == targetX) 
				{
					if (tempTargetY == targetY)
					{
						collection.add(po);
						collectionIterator.remove();	
					}
				} 		
			}
			
			// now for grouping
			
			// create group of elements
			// need to make sure we don't group too many packages...don't
			// want to exceed the weight limit...
			Group g = new Group();
			for (int i=0; i<collection.size(); i++)
			{
				PackageObject p = (PackageObject) collection.get(i);
				if (p.getWeight() + g.getWeight() < weightLimit)
				{
					g.addPackage(p);
				}
				else
				{
					result.add(g);
					g = new Group();
					g.addPackage(p);
				}
			}
			result.add(g);
		}	
	
		return result;	
	}
	
	protected String calculatePath()
	{
		String result = null;
		int closestPackageIndex = -1;
		int closestPackageDist = Integer.MAX_VALUE;
		int i = 0;
		System.out.println("OurBot<Seeking> calculating path to packages");
		
		// look through catalog to find close packages and keep closest one
		Vector packCatalog = m_gg.getPackageCatalog();
		for (i=0; i<packCatalog.size(); i++)
		{
			PackageObject po = (PackageObject) packCatalog.elementAt(i);
//			int dist = PackageObject.getShortestDistance(po, m_bot.getLocationX(), m_bot.getLocationY()); 
			int dist = 0;
			try
			{
				String route = m_gg.plotRouteToPoint(new Point(po.getLocationX(), po.getLocationY()), m_bot.getLocationX(), m_bot.getLocationY());
				StringTokenizer stp = new StringTokenizer(route, " ");
				dist = stp.countTokens();
				if (route.equals("X")) dist = 0;
			}
			catch (Exception pe)
			{
//				System.out.println(pe.getMessage());
//				pe.printStackTrace();
				dist = PackageObject.getShortestDistance(po, m_bot.getLocationX(), m_bot.getLocationY());
			}
			if ((dist < closestPackageDist) && (po.getState() != PackageObject.FUTILE_STATE))
			{
				boolean pointReject = false;
				Point pp = new Point(po.getLocationX(), po.getLocationY());
				for (int j=0; j<m_NoPackagesCollection.size(); j++)
				{
					Point tempPackagePoint = (Point) m_NoPackagesCollection.get(j);
					if (tempPackagePoint.x == pp.x)
					{
						if (tempPackagePoint.y == pp.y)
						{
							pointReject = true;
						}
					}
				}
				if (pointReject == false)
				{
					closestPackageIndex = i;
					closestPackageDist = dist;
				}	
			}
		}
		
		// now if closestPackageID != -1 then we have the closest package known
		// let's see if a home base is closer and if it has packages or not	
		Vector bases = m_gg.getHomeBases();
		int closestBaseIndex = -1;
		int closestBaseDistance = Integer.MAX_VALUE;
		for (i=0; i<bases.size(); i++)
		{
			Point pb = (Point) bases.get(i);
//			int baseDistance = computeDistance(pb, new Point(m_bot.getLocationX(), m_bot.getLocationY()));
			int baseDistance = 0;
			try
			{
				String route = m_gg.plotRouteToPoint(pb, m_bot.getLocationX(), m_bot.getLocationY());
//				System.out.println("route to base: " + route); 
				StringTokenizer st1 = new StringTokenizer(route, " ");
				baseDistance = st1.countTokens();
				if (route.equals("X"))
					baseDistance = 0;
			}
			catch (Exception be)
			{
//				System.out.println(be.getMessage());
//				be.printStackTrace();
				baseDistance = computeDistance(pb, new Point(m_bot.getLocationX(), m_bot.getLocationY()));
			}
			if (baseDistance < closestBaseDistance)
			{
				boolean baseReject = false;
				for (int k=0; k<m_NoPackagesCollection.size(); k++)
				{
					Point tempBasePoint = (Point) m_NoPackagesCollection.get(k);
					if (tempBasePoint.x == pb.x)
					{
						if (tempBasePoint.y == pb.y)
						{
							baseReject = true;
						}
					}
				}
				if (baseReject == false)
				{
					closestBaseIndex = i;
					closestBaseDistance = baseDistance;
				}	
			} 	
		}
		
		if ((closestPackageIndex == -1) && (closestBaseIndex == -1))
		{
			// hmmm...nothing found. Let's move randomly
//			System.out.println("OurBot<Seeking> failed to aquire any target");
			// maybe we can't find any more because some fell where we marked that there
			// weren't any...needed this for multiplayer servers during development.
			m_NoPackagesCollection = null;
			m_NoPackagesCollection = new Vector();
			
//			result = "N";	// don't pick N...check N, if valid, use, if not, choose E, then S, then W
			result = m_gg.pickSafeDirection(m_bot.getLocationX(), m_bot.getLocationY());
			return result;	
		}
		
		Point newDestination = null;
		
		if (closestBaseDistance <= closestPackageDist)	// let's favor bases as they tend to have more packages
		{
//			System.out.println("OurBot<Seeking> found base");
			newDestination = (Point) bases.get(closestBaseIndex);		
		}
		else
		{
//			System.out.println("OurBot<Seeking> found package");
			newDestination = new Point(((PackageObject)packCatalog.get(closestPackageIndex)).getLocationX(), ((PackageObject)packCatalog.get(closestPackageIndex)).getLocationX());;	
		}
		result = computeRouteToPoint(newDestination);
		
		return result;
	}
	
}